/*
 * peps.c - Pnm from EPS - v1.0
 *
 * by G. Adam Stanislav
 *
 * Started:	24-Jun-2001
 * Updated:	04-Jul-2001
 *
 * Copyright 2001 G. Adam Stanislav.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * This program converts EPS (encapsulated PostScript) files to
 * either PGM (portable graymap) or PPM (portable pixmap) files.
 *
 * It requires that gs (GhostScript) be installed and available
 * on the path.
 *
 * Options allow to use anti-aliasing setting of 1 (none), 2 (some),
 * or 4 (much) for the text and the graphics. The text anti-aliasing
 * is set by using the -t1, -t2, or -t4 options; the graphic
 * anti-aliasing with the -g1, -g2, and -g4 options. -t4 and -g4 are
 * the default.
 *
 * Secondly, options allow you to choose the resolution of the image
 * in pixels per inch (these are gs units, I would rather use metric
 * units, by gs uses inches). The default is 100 ppi. You can change
 * it with the -r switch, e.g., -r144 will increase the resolution.
 *
 * If you need a different vertical and horizontal resolution, use
 * -v and -h instead, e.g., -v1200 -h600. This may be necessary with
 * certain printers.
 *
 * All output is sent to stdout (directly by gs), from where it can
 * be piped or redirected. The idea is to use it with image format
 * converters. Both, PGM and PPM, are text files, so they can grow
 * huge. However, using this format is the only way to get gs to
 * use anti-aliasing, at least as of this writing (24-Jun-2001).
 */
#include <stdio.h>
#include <stdlib.h>

char gsargs[] = "gs "
	"-q "
	"-dBATCH "
	"-dGraphicsAlphaBits=%i "
	"-dTextAlphaBits=%i "
	"-r%ix%i "
	"-g%ix%i "
	"-dNOPAUSE "
	"%s"
	"-sDEVICE=pnm "
	"-sOutputFile=- "
	"-"
	"\0| pnmtopng";

char *input  = NULL;
char *output = NULL;

int hres = 100;
int vres = 100;
int gab  = 4;
int tab  = 4;

int png  = 0;

int  x = 0;
int  y = 0;
int dx = 0;
int dy = 0;
int dX = 0;
int dY = 0;

int width  = -1;
int height = -1;

int angle  =  0;

char safer[] = "-dSAFER ";
char buffer[1024];

int usage(void) {
	fprintf(stderr, "Usage: peps [option ...] file\n"
		"\t-g<b>\t-t<b>\n"
		"\t-r<r>\t-h<r>\n"
		"\t-v<r>\t-a<d>\n"
		"\t-x<p>\t-X<p>\n"
		"\t-y<p>\t-Y<p>\n"
		"\t-s\t-p\n"
		"\t-o file\n"
		"\t<b> = 1 | 2 | 4\n"
		"\t<r> = resolution in ppi\n"
		"\t<d> = angle in degrees\n"
		"\t<p> = offset in points\n");
	return 1;
}

int invalid(FILE *eps) {
	fprintf(stderr, "peps: %s is not a valid EPS file.\n", input);
	if (eps) fclose(eps);
	return 4;
}

int parse (FILE *eps) {
	register char *charptr;

	while (!feof(eps)) {
		fgets(buffer, 1024, eps);
		if (!strncmp(buffer, "%%BoundingBox:", 14)) {
			charptr = buffer + 14;
			while (*charptr <= ' ') {
				if (!*charptr) break;
				charptr++;
			}

			if (((*charptr == '-') && (isdigit(charptr[1]))) ||
				isdigit(*charptr)) {
				x = -atoi(charptr);

				while (*charptr > ' ') charptr++;
				while (*charptr <= ' ') {
					if (!*charptr) break;
					charptr++;
				}

				if (((*charptr == '-') && (isdigit(charptr[1]))) ||
					isdigit(*charptr)) {
					y = -atoi(charptr);

					while (*charptr > ' ') charptr++;
					while (*charptr <= ' ') {
						if (!*charptr) break;
						charptr++;
					}

					if (((*charptr == '-') && (isdigit(charptr[1]))) ||
						isdigit(*charptr)) {
						width = atoi(charptr) + x;

						while (*charptr > ' ') charptr++;
						while (*charptr <= ' ') {
							if (!*charptr) break;
							charptr++;
						}

						if (((*charptr == '-') && (isdigit(charptr[1]))) ||
							isdigit(*charptr)) {
							height = atoi(charptr) + y;
							if ((width > 0) && (height > 0))
								return 0;
						}
					}
				}
			}
		}
	}
	return 1;
}

int main(int argc, char *argv[]) {
	register int i;
	register char *charptr;
	FILE *gs, *eps;

	if (argc < 2) return usage();

	for (i = 1; i < argc; i++) {
		charptr = argv[i];

		if (*charptr == '-') {
			charptr++;
			switch (*charptr) {
			default:
				return usage();
			case 'g':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (!charptr || charptr[1])
					return usage();
				switch (*charptr) {
				case '1':
					gab = 1;
					break;
				case '2':
					gab = 2;
					break;
				case '4':
					gab = 4;
					break;
				default:
					return usage();
				}
				break;
			case 't':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (!charptr || charptr[1])
					return usage();
				switch (*charptr) {
				case '1':
					tab = 1;
					break;
				case '2':
					tab = 2;
					break;
				case '4':
					tab = 4;
					break;
				default:
					return usage();
				}
				break;
			case 'r':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && isdigit(*charptr))
					hres = vres = atoi(charptr);
				else return usage();
				if (!hres) {
					fprintf(stderr, "peps: Resolution cannot be 0.\n");
					return usage();
				}
				break;
			case 'v':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && isdigit(*charptr))
					vres = atoi(charptr);
				else return usage();
				if (!vres) {
					fprintf(stderr, "peps: Resolution cannot be 0.\n");
					return usage();
				}
				break;
			case 'h':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && isdigit(*charptr))
					hres = atoi(charptr);
				else return usage();
				if (!hres) {
					fprintf(stderr, "peps: Resolution cannot be 0.\n");
					return usage();
				}
				break;
			case 's':
				if (charptr[1] == '\0')
					safer[0] = '\0';
				else return usage();
				break;
			case 'p':
				if (charptr[1] == '\0')
					png = 1;
				else return usage();
				break;
			case 'o':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (!charptr) return usage();
				output = charptr;
				break;
			case 'x':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && (isdigit(*charptr) ||
				((*charptr == '-') && isdigit(charptr[1]))))
					dx = atoi(charptr);
				else return usage();
				break;
			case 'y':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && (isdigit(*charptr) ||
				((*charptr == '-') && isdigit(charptr[1]))))
					dy = atoi(charptr);
				else return usage();
				break;
			case 'X':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && (isdigit(*charptr) ||
				((*charptr == '-') && isdigit(charptr[1]))))
					dX = atoi(charptr);
				else return usage();
				break;
			case 'Y':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && (isdigit(*charptr) ||
				((*charptr == '-') && isdigit(charptr[1]))))
					dY = atoi(charptr);
				else return usage();
				break;
			case 'a':
				charptr++;
				if (!*charptr) charptr = argv[++i];
				if (charptr && (isdigit(*charptr) ||
				((*charptr == '-') && isdigit(charptr[1]))))
					angle = atoi(charptr);
				else return usage();
				break;				
			}
		}
		else input = charptr;
	}

	if (!input) return usage();

	if (output && (freopen(output, "w", stdout) == NULL)) {
		perror(output);
		return 7;
	}

	if ((eps = fopen(input, "r")) == NULL) {
		perror(input);
		return 3;
	}

	if (png) gsargs[strlen(gsargs)] = ' ';

	/* Parse input file */
	fgets(buffer, 1024, eps);
	if ((strncmp(buffer, "%!PS-Adobe-", 11)) || parse(eps))
		return invalid(eps);

	fclose(eps);

	x += dx;
	y += dy;
	width  += dx + dX;
	height += dy + dY;

	if ((width <= 0) || (height <= 0)) {
		fprintf(stderr, "peps: Image absorbed by black hole.\n");
		return 13;
	}

	sprintf(buffer, gsargs,
		gab,
		tab,
		hres, vres,
		(int)(((double)hres * (double)width  / 72.0) + 0.5),
		(int)(((double)vres * (double)height / 72.0) + 0.5),
		safer);

	if ((gs = popen(buffer, "w")) == NULL) {
		perror("peps (gs)");
		return 5;
	}

	fprintf(gs,
		"%%!PS-Adobe-3.0\n"
		"%%%%BoundingBox: 0 0 %i %i\n"
		"%%%%Creator: peps-1.0\n"
		"%%%%EndComments\n"
		"%%%%BeginProlog\n"
		"/BeginEPSF {\n"
		"/peps_state save def\n"
		"/dict_count countdictstack def\n"
		"/op_count count 1 sub def\n"
		"userdict begin\n"
		"/showpage { } def\n"
		"0 setgray 0 setlinecap\n"
		"1 setlinewidth 0 setlinejoin\n"
		"10 setmiterlimit [ ] 0 setdash newpath\n"
		"/languagelevel where\n"
		"{pop languagelevel\n"
		"1 ne\n"
		"{false setstrokeadjust false setoverprint\n"
		"} if } if } bind def\n\n"
		"/EndEPSF {\n"
		"count op_count sub {pop} repeat\n"
		"countdictstack dict_count sub {end} repeat\n"
		"peps_state restore\n"
		"} bind def\n"
		"%%%%EndProlog\n"
		"%%%%BeginSetup\n"
		"%%%%EndSetup\n"
		"%%%%Page: 1 1\n"
		"%%%%BeginPageSetup\n"
		"%%%%EndPageSetup\n"
		"BeginEPSF\n"
		"%i %i 2 div add %i %i 2 div add translate\n"
		"%i rotate\n"
		"%i 2 div %i 2 div translate\n"
		"%%%%BeginDocument: %s\n"
		"(%s) run\n"
		"%%%%EndDocument\n"
		"EndEPSF\n"
		"showpage\n"
		"%%%%EOF\n"
		"quit\n",
		width, height,
		x, width, y, height,
		angle,
		-width, -height,
		input, input
		);

	pclose(gs);

	return 0;
}

